package net.h34t.filemure

import io.javalin.Javalin
import io.javalin.http.UnauthorizedResponse
import net.h34t.filemure.classification.AIClassifier
import net.h34t.filemure.controller.*
import net.h34t.filemure.repository.SqliteRepository
import net.h34t.filemure.tpl.Frame
import net.h34t.filemure.tpl.Unauthorized

class FilemureApp(
    repository: SqliteRepository,
    contentExtractor: ContentExtractor) {

    private val modifiers = TemplateModifiers()

    private val loginPageController = AccountController(modifiers, repository)

    private val overviewController = OverviewController(modifiers, repository)
    private val limboController = LimboController(modifiers, repository)
    private val uploadController = UploadController(modifiers, repository, contentExtractor)
    private val documentController = DocumentController(modifiers, repository)
    private val searchController = SearchController(modifiers, repository)

    fun register(server: Javalin) {
        server.beforeMatched { ctx ->
            val userRole = getUserRole(ctx)
            if (ctx.routeRoles().isNotEmpty() && !ctx.routeRoles().contains(userRole)) {
                ctx.redirectPRG("/login")
                throw UnauthorizedResponse()
            }
        }

        server.get("/login", loginPageController::formLogin, Role.ANON, Role.USER)
        server.post("/login", loginPageController::doLogin, Role.ANON, Role.USER)
        server.post("/logout", loginPageController::doLogout, Role.USER)

        server.get("/register", loginPageController::formRegister, Role.ANON, Role.USER)
        server.post("/register", loginPageController::doRegister, Role.ANON, Role.USER)

        server.post("/upload", uploadController::upload, Role.USER)
        server.get("/limbo", limboController::formLimbo, Role.USER)

        server.get("/", overviewController::overview, Role.USER)
        server.get("/overview/{year}/{month}", overviewController::overviewCategory, Role.USER)

        server.get("/document/new", documentController::createDocumentForm, Role.USER)
        server.post("/document/new", documentController::createDocument, Role.USER)
        server.get("/document/{extId}", documentController::documentDetail, Role.USER)
        server.get("/document/{extId}/download", documentController::downloadDocument, Role.USER)
        server.get("/document/{extId}/edit", documentController::editDocumentForm, Role.USER)
        server.post("/document/{extId}/edit", documentController::editDocumentAction, Role.USER)
        server.post("/document/{extId}/delete", documentController::deleteDocumentAction, Role.USER)

        server.get("/file/{extId}/download", documentController::downloadFile, Role.USER)
        server.post("/file/{extId}/delete", documentController::deleteFileAction, Role.USER)

        server.get("/search", searchController::search, Role.USER)
        server.get("/tags", searchController::tags, Role.USER)

        server.exception(UnauthorizedResponse::class.java) { e, ctx ->
            ctx.tempolin(
                Frame(
                    modifiers = modifiers,
                    title = "unauthorized",
                    target = "",
                    content = Unauthorized(),
                    back = "",
                    logout = false
                )
            )
        }
    }
}
