package net.h34t.filemure

import io.javalin.Javalin
import io.javalin.http.staticfiles.Location
import net.h34t.filemure.classification.AIClassifier
import net.h34t.filemure.repository.SqliteRepository
import org.eclipse.jetty.http.HttpCookie
import org.eclipse.jetty.server.session.DefaultSessionCache
import org.eclipse.jetty.server.session.FileSessionDataStore
import org.eclipse.jetty.server.session.SessionHandler
import java.io.File
import java.time.Duration
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter

fun main() {
    val dtf = DateTimeFormatter.ISO_DATE_TIME

    val db = System.getenv("db_path")
        ?: throw IllegalArgumentException("Please define an env dbpath, e.g. /data/filemure.db")

    val sessionExpiry = System.getenv("session_expiry_sec")?.toInt() ?: (Duration.ofMinutes(30).toSeconds().toInt())

    val aiClassifier = AIClassifier(
        key = System.getenv("ai_key")
    )

    val contentExtractor = ContentExtractor(aiClassifier)

    val app = FilemureApp(SqliteRepository("jdbc:sqlite:$db"), contentExtractor)

    Javalin
        .create { config ->
            config.staticFiles.add { staticFiles ->
                staticFiles.hostedPath = "/"
                staticFiles.directory = "public"
                staticFiles.location = Location.CLASSPATH
                staticFiles.precompress = true
            }

            config.requestLogger.http { ctx, ms ->
                println("${dtf.format(LocalDateTime.now())}\t${ctx.method()}\t${ctx.path()}\t$ms")
            }

            config.useVirtualThreads = true

            config.jetty.modifyServletContextHandler {
                it.sessionHandler = fileSessionHandler(sessionExpiry)
            }
        }
        .also {
            app.register(it)
        }
        .start(7070)
}

fun fileSessionHandler(expirySec: Int) = SessionHandler().apply {
    sessionCache = DefaultSessionCache(this).apply {
        sessionDataStore = FileSessionDataStore().apply {
            val baseDir = File(System.getProperty("java.io.tmpdir"))
            this.storeDir = File(baseDir, "javalin-session").apply { mkdirs() }
        }
    }
    maxInactiveInterval = expirySec
    httpOnly = true
    isSecureRequestOnly = true
    sameSite = HttpCookie.SameSite.STRICT
}