package net.h34t.filemure.classification

import com.fasterxml.jackson.annotation.JsonProperty
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import okhttp3.MediaType.Companion.toMediaType
import okhttp3.OkHttpClient
import okhttp3.Request
import okhttp3.RequestBody
import okhttp3.RequestBody.Companion.toRequestBody
import java.util.*

/**
 * examples for models:
 * Vision: "meta-llama/llama-3.2-11b-vision-instruct"
 * Text: "meta-llama/llama-3.2-3b-instruct"
 */
class AIClassifier(
    val key: String,
    val visionModel: String = "meta-llama/llama-3.2-90b-vision-instruct",
    val textModel: String = "meta-llama/llama-3.3-70b-instruct",
    val baseUrl: String = "https://openrouter.ai/api/v1/chat/completions"
) {

    data class OpenrouterRequest(
        val model: String,
        val messages: List<ReqMessage>
    )

    interface ContentType

    data class ReqMessage(
        val role: String,
        val content: List<ContentType>
    )

    data class TextContent(
        val type: String = "text",
        val text: String
    ) : ContentType

    data class ImageContent(
        val type: String = "image_url",
        val image_url: ImageUrl
    ) : ContentType

    data class ImageUrl(
        val url: String
    )

    data class OpenrouterResponse(
        @JsonProperty("id")
        val id: String,
        @JsonProperty("provider")
        val provider: String,
        @JsonProperty("model")
        val model: String,
        @JsonProperty("object")
        val contentObject: String,
        @JsonProperty("created")
        val created: Long,
        @JsonProperty("choices")
        val choices: List<Choice>,
        @JsonProperty("system_fingerprint")
        val systemFingerprint: String?,
        @JsonProperty("usage")
        val usage: Usage
    )

    data class Choice(
        @JsonProperty("logprobs")
        val logprobs: Unit?,
        @JsonProperty("finish_reason")
        val finishReason: String?,
        @JsonProperty("native_finish_reason")
        val nativeFinishReason: String?,
        @JsonProperty("index")
        val index: Int,
        @JsonProperty("message")
        val message: ResMessage
    )

    data class ResMessage(
        @JsonProperty("role")
        val role: String,
        @JsonProperty("content")
        val content: String,
        @JsonProperty("refusal")
        val refusal: String?
    )

    data class Usage(
        @JsonProperty("prompt_tokens")
        val prompt_tokens: Int,
        @JsonProperty("completion_tokens")
        val completion_tokens: Int,
        @JsonProperty("total_tokens")
        val total_tokens: Int
    )

    private val encoder = Base64.getEncoder()

    fun queryText(
        query: String,
        model: String = textModel,
    ): String {
        return query(
            model = model,
            content = listOf(TextContent(text = query))
        )
    }

    private fun query(content: List<ContentType>, model: String): String {
        val orRequest = OpenrouterRequest(
            model = model,
            messages = listOf(
                ReqMessage(
                    role = "user",
                    content = content
                )
            )
        )

        val mapper = ObjectMapper()
        val bodyContent = mapper.writeValueAsString(orRequest)

        val client = OkHttpClient()

        val json = "application/json".toMediaType()

        val body: RequestBody = bodyContent.toRequestBody(json)
        val req = Request.Builder()
            .url(baseUrl)
            .header("Authorization", "Bearer $key")
            .header("Content-Type", "application/json")
            .post(body)
            .build()

        client.newCall(req).execute().use { res ->
            val bodyString = res.body?.string()?.trim()
            val responseBody = bodyString?.let { mapper.readValue<OpenrouterResponse>(it) }
                ?: throw Exception("body is null")

            return responseBody.choices.joinToString("\n") { it.message.content }.trim()
        }
    }

    fun classifyImage(
        prompt: String,
        contentType: String,
        data: ByteArray,
        model: String = visionModel
    ): String {
        val dataUrl = "data:$contentType;base64," + encoder.encodeToString(data)

        return query(
            model = model,
            content = listOf(
                TextContent(
                    type = "text",
                    // text = "Extract the message text only in this file and repeat it word-for-word."
                    text = prompt
                ),
                ImageContent(
                    type = "image_url",
                    image_url = ImageUrl(url = dataUrl)
                )
            )
        )
    }
}