package net.h34t.filemure.repository

import app.cash.sqldelight.ColumnAdapter
import app.cash.sqldelight.db.SqlDriver
import app.cash.sqldelight.driver.jdbc.sqlite.JdbcSqliteDriver
import at.favre.lib.crypto.bcrypt.BCrypt
import net.h34t.filemure.*
import net.h34t.filemure.TagAdapter.serialize
import net.h34t.filemure.db.Account
import net.h34t.filemure.db.Database
import net.h34t.filemure.db.File_
import java.io.InputStream
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter

class SqliteRepository(url: String) {

    private companion object Adapters {
        private val sqliteDtf = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")

        val localDateTimeAdapter = object : ColumnAdapter<LocalDateTime, String> {
            override fun decode(databaseValue: String): LocalDateTime {
                return LocalDateTime.parse(databaseValue, sqliteDtf)
            }

            override fun encode(value: LocalDateTime): String {
                return value.format(sqliteDtf)
            }
        }

        val stateAdapter = object : ColumnAdapter<State, Long> {
            override fun decode(databaseValue: Long): State = State.fromCode(databaseValue.toInt())
            override fun encode(value: State): Long = value.code.toLong()
        }

        val tagsAdapter = object : ColumnAdapter<List<Tag>, String> {
            override fun decode(databaseValue: String): List<Tag> = TagAdapter.parse(databaseValue)
            override fun encode(value: List<Tag>): String = value.serialize()
        }

        val extIdAdapter = object : ColumnAdapter<ExtId, String> {
            override fun decode(databaseValue: String): ExtId = ExtId(databaseValue)
            override fun encode(value: ExtId): String = value.value
        }
    }

    private val database: Database

    init {
        val driver: SqlDriver = JdbcSqliteDriver(url)

        try {
            Database.Schema.create(driver)
        } catch (_: Exception) {
        }

        database = Database(
            driver = driver,
            accountAdapter = Account.Adapter(
                ext_idAdapter = extIdAdapter,
                createdAdapter = localDateTimeAdapter,
                stateAdapter = stateAdapter
            ),
            documentAdapter = net.h34t.filemure.db.Document.Adapter(
                tagsAdapter = tagsAdapter,
                createdAdapter = localDateTimeAdapter,
                reference_dateAdapter = localDateTimeAdapter,
                stateAdapter = stateAdapter,
                ext_idAdapter = extIdAdapter
            ),
            file_Adapter = File_.Adapter(
                createdAdapter = localDateTimeAdapter,
                stateAdapter = stateAdapter,
                ext_idAdapter = extIdAdapter
            )
        )
    }

    private val bcrypt = BCrypt.withDefaults()

    private fun generateExtId(): ExtId = ExtId.generate()

    fun addAccount(
        email: String,
        password: CharArray
    ) {
        val bcryptPassword = bcrypt.hashToString(6, password)
        val extId = generateExtId()
        database.databaseQueries.addAccount(
            extId = extId,
            email = email,
            password = bcryptPassword,
            state = State.ACTIVE
        )
    }

    fun loginAccount(email: String, password: CharArray): AccountRef? {
        return database.databaseQueries.loginAccount(email).executeAsList().firstOrNull()?.let { account ->
            if (account.state != State.ACTIVE)
                return null

            val verify = BCrypt.verifyer().verify(password, account.password)

            if (verify.verified) {
                account.let {
                    AccountRef(
                        id = it.id,
                        extId = it.ext_id,
                        email = it.email,
                        created = it.created,
                        state = it.state
                    )
                }
            } else {
                null
            }
        }
    }

    fun addFileToLimbo(
        accountId: Long,
        filename: String,
        contentType: String?,
        size: Long,
        contentExtracted: String,
        content: ByteArray
    ): IdPair =
        database.databaseQueries.transactionWithResult {
            val extId = generateExtId()
            database.databaseQueries.insertFileIntoLimbo(
                account_id = accountId,
                ext_id = extId,
                filename = filename,
                content_type = contentType,
                file_size = size,
                content_extracted = contentExtracted,
                content = content
            )
            IdPair(
                id = lastInsertedId(),
                extId = extId
            )
        }

    fun addNewFileToDocument(
        accountId: Long,
        documentId: Long,
        filename: String,
        contentType: String?,
        size: Long,
        content: InputStream
    ) =
        database.databaseQueries.transactionWithResult {
            generateExtId().let { extId ->
                database.databaseQueries.insertFileForDocument(
                    account_id = accountId,
                    document_id = documentId,
                    ext_id = extId,
                    filename = filename,
                    content_type = contentType,
                    file_size = size,
                    content = content.readAllBytes()
                )
                IdPair(id = lastInsertedId(), extId = extId)
            }
        }

    fun getLimboFileCount(accountId: Long, state: State = State.ACTIVE): Long {
        return database.databaseQueries.getLimboFileCount(account_id = accountId, state = state)
            .executeAsOne()
    }

    fun getFilesInLimbo(accountId: Long, state: State = State.ACTIVE): List<FileRef> {
        return database.databaseQueries.getFilesInLimbo(account_id = accountId, state = state)
            .executeAsList()
            .map {
                FileRef(
                    id = it.id,
                    accountId = it.account_id,
                    extId = it.ext_id,
                    documentId = null,
                    filename = it.filename,
                    fileSize = it.file_size,
                    contentType = it.content_type,
                    contentExtracted = it.content_extracted,
                    created = it.created,
                    state = it.state
                )
            }
    }

    fun addDocument(
        accountId: Long,
        title: String,
        referenceDate: LocalDateTime,
        tags: List<Tag>,
        description: String,
        fileExtIds: List<ExtId>
    ): ExtId {
        val extId = generateExtId()
        database.databaseQueries.transaction {
            database.databaseQueries.addDocument(
                account_id = accountId,
                ext_id = extId,
                title = title,
                description = description,
                tags = tags,
                reference_date = referenceDate
            )

            val documentId = database.databaseQueries.getLastInsertRowId().executeAsOne()

            database.databaseQueries.attachLimboFilesToDocument(
                account_id = accountId,
                document_id = documentId,
                ext_id = fileExtIds
            )
        }

        return extId;
    }

    fun getDocuments(accountId: Long, state: State = State.ACTIVE): List<Document> {
        return database.databaseQueries.getDocuments(account_id = accountId, state = state)
            .executeAsList()
            .map {
                Document(
                    id = it.id,
                    extId = it.ext_id,
                    title = it.title,
                    description = it.description,
                    tags = it.tags,
                    created = it.created,
                    referenceDate = it.reference_date,
                    state = it.state,
                    files = emptyList()
                )
            }
    }

    fun getDocumentsByYearMonth(accountId: Long, year: Int, month: Int, state: State = State.ACTIVE) =
        "%04d%02d".format(year, month).let { date ->
            database.databaseQueries.getDocumentsByYearMonth(
                account_id = accountId,
                yearMonth = date,
                state = state
            ).executeAsList().map {
                Document(
                    id = it.id,
                    extId = it.ext_id,
                    title = it.title,
                    description = it.description,
                    tags = it.tags,
                    created = it.created,
                    referenceDate = it.reference_date,
                    state = it.state,
                    files = emptyList()
                )
            }
        }

    fun updateDocument(
        accountId: Long,
        id: Long,
        title: String,
        referenceDate: LocalDateTime,
        tags: List<Tag>,
        description: String,
        state: State = State.ACTIVE,
    ) {
        database.databaseQueries.updateDocument(
            id = id,
            title = title,
            reference_date = referenceDate,
            tags = tags,
            description = description,
            state = state,
            account_id = accountId,
        )
    }

    fun getDocumentByExtId(accountId: Long, extId: ExtId, state: State): Document {
        return database.databaseQueries.getDocumentByExtId(
            account_id = accountId,
            ext_id = extId,
            state = state
        ).executeAsOne().let { d ->
            Document(
                id = d.id,
                extId = d.ext_id,
                title = d.title,
                description = d.description,
                tags = d.tags,
                created = d.created,
                referenceDate = d.reference_date,
                state = d.state,
                files = database.databaseQueries.getFilesForDocument(
                    document_id = d.id,
                    account_id = accountId
                ).executeAsList().map { f ->
                    FileRef(
                        id = f.id,
                        extId = f.ext_id,
                        accountId = f.account_id,
                        documentId = f.document_id,
                        filename = f.filename,
                        contentType = f.content_type,
                        contentExtracted = f.content_extracted,
                        fileSize = f.file_size,
                        created = f.created,
                        state = f.state
                    )
                }
            )

        }
    }

    fun searchDocuments(accountId: Long, state: State = State.ACTIVE, query: String): List<Document> {
        return database.databaseQueries.searchDocuments(account_id = accountId, state = state, query = query)
            .executeAsList()
            .map {
                Document(
                    id = it.id,
                    extId = it.ext_id,
                    title = it.title,
                    description = it.description,
                    tags = it.tags,
                    created = it.created,
                    referenceDate = it.reference_date,
                    state = it.state,
                    files = emptyList()
                )
            }
    }

    fun loadFile(accountId: Long, extId: ExtId): FileContent {
        return database
            .databaseQueries
            .getFile(account_id = accountId, ext_id = extId)
            .executeAsOne().let { f ->
                FileContent(
                    id = f.id,
                    extId = f.ext_id,
                    filename = f.filename,
                    contentType = f.content_type,
                    contentExtracted = f.content_extracted,
                    fileSize = f.file_size,
                    content = f.content
                )
            }
    }

    fun setDocumentState(accountId: Long, extId: ExtId, state: State) {
        database.databaseQueries.setDocumentState(account_id = accountId, ext_id = extId, state = state)
    }

    fun setFileState(accountId: Long, extId: ExtId, state: State) {
        database.databaseQueries.setFileState(account_id = accountId, ext_id = extId, state = state)
    }

    fun getAllTags(accountId: Long): List<Tag> =
        database.databaseQueries.getAllTags(accountId).executeAsList().flatten().distinct()


    private fun lastInsertedId() = database.databaseQueries.getLastInsertRowId().executeAsOne()

    fun getDocumentsByTags(accountId: Long, state: State = State.ACTIVE, tags: List<Tag>): List<Document> {
        return database.databaseQueries.getDocuments(account_id = accountId, state = state)
            .executeAsList()
            .filter { d -> (d.tags intersect tags).isNotEmpty() }
            .map {
                Document(
                    id = it.id,
                    extId = it.ext_id,
                    title = it.title,
                    description = it.description,
                    tags = it.tags,
                    created = it.created,
                    referenceDate = it.reference_date,
                    state = it.state,
                    files = emptyList()
                )
            }
    }
}